#!/usr/bin/python


from .jni import *
import subprocess
import re


__all__ = ['JavaP']


class JavaP(object):
    
    javap = '/usr/bin/javap'

    # more specific regex
    re_method = re.compile(r'''(?:\ \ )
        (?:(?P<access>public|private|protected)\ )?
        (?P<modifiers>(?:(?:static|final|abstract|synchronized|native|transient)\ )*)
        (?P<restype>[\w., \$\?<>\[\]]+)(?:\ )
        (?P<name>[\w\$]+)
        \(
        (?P<argtypes>[^)]*)
        \).*;''', re.X)
    re_field = re.compile(r'''(?:\ \ )
        (?:(?P<access>public|private|protected)\ )?
        (?P<modifiers>(?:(?:static|final|abstract|synchronized|native|transient|volatile)\ )*)
        (?P<type>[\w., \$\?<>\[\]]+)(?:\ )
        (?P<name>[\w\$]+);''', re.X)
    re_descriptor = re.compile(r'''(?:\ \ \ \ descriptor:\ )
        (?P<signature>.*)''', re.X)

    @classmethod
    def parse_class(cls, classname, classpath=None):
        cmd = [cls.javap, '-p', '-s']
        classpath = classpath or []
        for p in classpath:
            cmd.extend(['-cp', p])
        cmd.append(classname)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        clazz = JniClass(classname)

        while True:
            line = p.stdout.readline().decode('utf-8')
            if not line:
                break

            # end of parsing (static initializer block or end of class)
            if line.startswith('  static {') or line.startswith('}'):
                break

            # skip blank lines and first line
            line = line.rstrip()
            if not line or line.startswith('Compiled'):
                continue

            # class
            if line.startswith('public class'):
                name = line.split()[2]
                angle = name.find('<')
                if angle != -1:
                    name = name[:angle]
                if name != clazz.name:
                    raise IOError('Bad Class name')

            # this shouldn't happen
            elif line.startswith('    '):
                raise IOError('Unexpected line')

            # constructor, method or field
            elif line.startswith('  '):

                member_cls = None
                name = signature = None
                kwargs = {}

                # constructor or method
                if '(' in line and ')' in line:

                    # constructor
                    if clazz.name + '(' in line:
                        member_cls = JniConstructor
                        name = clazz.shortname

                    # method
                    else:
                        match = cls.re_method.match(line)
                        if not match:
                            raise IOError('Failed to match method: %s' % line)
                        member_cls = JniMethod
                        name = match.group('name')
                        if 'static' in match.group('modifiers'):
                            kwargs['flags'] = JniMethod.FLAG_STATIC

                # field
                else:
                    match = cls.re_field.match(line)
                    if not match:
                        raise IOError('Failed to match field: %s' % line)  
                    member_cls = JniField
                    name = match.group('name')
                    if 'static' in match.group('modifiers'):
                        kwargs['flags'] = JniField.FLAG_STATIC

                # descriptor
                line = p.stdout.readline().decode('utf-8')
                if not line:
                    raise IOError('Missing descriptor')
                line = line.rstrip()
                match = cls.re_descriptor.match(line)
                if not match:
                    raise IOError('Failed to match descriptor')
                signature = match.group('signature')

                # add member to class
                if member_cls is None:
                    raise IOError('Failed to match class member')
                member_cls(name, signature, **kwargs).add_to(clazz)
    
        # wait for process to end
        out, err = p.communicate()
        if p.returncode != 0:
            raise IOError(str(err).rstrip())
        return clazz


if __name__ == '__main__':
    import sys
    classpath = sys.argv[1]
    classname = sys.argv[2] if len(sys.argv) > 2 else None
    print(JavaP.parse_class(classname, classpath))



import SCons.Tool
import SCons.Util
import os


__all__ = [
    'exists',
    'generate',
]


NjiAction = SCons.Action.Action('$NJICCOM', '$NJICCOMSTR')


def NjiEmitter(target, source, env):
    targetDir, targetFile = os.path.split(SCons.Util.to_String(target[0]))
    targetName, targetExt = os.path.splitext(targetFile)
    if env['NJICPATHH']:
        targetDir = SCons.Util.to_String(env['NJICPATHH'])
    target.append(os.path.join(targetDir, targetName) + '.h')
    if env['NJICTEMPLATEC']:
        env.Depends(target[0], env['NJICTEMPLATEC'])
    if env['NJICTEMPLATEH']:
        env.Depends(target[1], env['NJICTEMPLATEH'])
    return target, source


class pathopt(object):
    """Copied from javac tool"""
    def __init__(self, opt, var, default=None):
        self.opt = opt
        self.var = var
        self.default = default

    def __call__(self, target, source, env, for_signature):
        path = env[self.var]
        if path and not SCons.Util.is_List(path):
            path = [path]
        if self.default:
            default = env[self.default]
            if default:
                if not SCons.Util.is_List(default):
                    default = [default]
                path = path + default
        if path:
            return [self.opt, os.pathsep.join(map(str, path))]
        else:
            return []


def generate(env):
    c_file, cxx_file = SCons.Tool.createCFileBuilders(env)
    c_file.add_action('.nji', NjiAction)
    c_file.add_emitter('.nji', NjiEmitter)
    env['NJIC'] = 'njic'
    env['NJICFLAGS'] = SCons.Util.CLVar('')
    env['NJICPATHH'] = None
    env['NJICTEMPLATEC'] = None
    env['NJICTEMPLATEH'] = None
    env['NJICCLASSPATH'] = []
    env['_njicpathopt'] = pathopt
    env['_NJICTEMPLATEC'] = '${_njicpathopt("--template-c", "NJICTEMPLATEC")} '
    env['_NJICTEMPLATEH'] = '${_njicpathopt("--template-h", "NJICTEMPLATEH")} '
    env['_NJICCLASSPATH'] = '${_njicpathopt("--classpath", "NJICCLASSPATH", "JAVACLASSPATH")} '
    env['NJICCOM'] = '$NJIC $NJICFLAGS $_NJICTEMPLATEC $_NJICTEMPLATEH $_NJICCLASSPATH --output-c ${TARGETS[0]} --output-h ${TARGETS[1]} $SOURCE'


def exists(env):
    return env.Detect(['njic'])



#!/usr/bin/python


import re


__all__ = ['JniClass', 'JniConstructor', 'JniField', 'JniMethod', 'JniAll']


JniAll = object()


class JniClass(object):

    SKIP_FIELDS = 1
    SKIP_CONSTRUCTORS = 2
    SKIP_METHODS = 4

    def __init__(self, name, altname=None, constructors=None, fields=None, methods=None, force=False):
        self.name = self.normalise(name)
        self.altname = altname
        self.fields = fields or []
        self.constructors = constructors or []
        self.methods = methods or []
        self.force = force

    def __str__(self):
        s = [self.name]
        for f in self.fields:
            s.append('  %s' % f)
        for m in self.methods:
            s.append('  %s' % m)
        return '\n'.join(s)

    @property
    def shortname(self):
        return self.altname or self.name.rsplit('.', 1)[-1].replace('$', '_')

    @property
    def cname(self):
        return self.name.replace('.', '_').replace('$', '_')

    @property
    def jname(self):
        return self.name.replace('.', '/')

    @staticmethod
    def normalise(name):
        if '/' in name:
            return '.'.join(name.split('/'))
        return name

    def add_field(self, field):
        self.fields.append(field)

    def add_constructor(self, constructor):
        self.constructors.append(constructor)

    def add_method(self, method):
        self.methods.append(method)

    def _reduce_members(self, javap, nji):
        if nji is JniAll:
            for j in javap:
                j.add_to(self)
            return
        for n in nji:
            found = 0
            for j in javap:
                if j == n:
                    j.reduce(j, n).add_to(self)
                    found += 1
            if not found:
                if n.force:
                    n.add_to(self)
                else:
                    raise ValueError("Cannot find %s" % n)

    @classmethod
    def reduce(cls, javap, nji):
        if javap.name != nji.name:
            raise ValueError("Classes don't match")

        # new reduced class
        new = cls(javap.name, altname=nji.altname)
        new._reduce_members(javap.fields, nji.fields)
        new._reduce_members(javap.constructors, nji.constructors)
        new._reduce_members(javap.methods, nji.methods)
        return new

    def validate(self):
        if not self.name:
            raise ValueError("Must have name")
        for field in self.fields:
            field.validate()
        for constructor in self.constructors:
            constructor.validate()
        for method in self.methods:
            method.validate()

    def uniqify(self):
        members = self.fields + self.constructors + self.methods
        while members:
            i, m0 = 0, members.pop(0)
            for m1 in members:
                if m0.cname == m1.cname:
                    i += 1
                    m1.altname = m1.cname + 'V%d' % i


JniAll = object()
"""Sentinel for 'all' members"""


class JniMember(object):

    FLAG_STATIC = 1
    FLAG_OPTIONAL = 2

    re_sig = re.compile(r'(?:\((?P<argtypes>[^)]*)\))?(?P<restype>.*)')
    re_type = re.compile(r'\[*(?:[ZBCSIJFDV]|L[^;]+;)')

    types = {
        'Z': 'jboolean',
        'B': 'jbyte',
        'C': 'jchar',
        'S': 'jshort',
        'I': 'jint',
        'J': 'jlong',
        'F': 'jfloat',
        'D': 'jdouble',
        'L': 'jobject',
        'V': 'void',
    }

    type_modifiers = {
        'jboolean': 'Boolean',
        'jbyte': 'Byte',
        'jchar': 'Char',
        'jshort': 'Short',
        'jint': 'Int',
        'jlong': 'Long',
        'jfloat': 'Float',
        'jdouble': 'Double',
        'void': 'Void',
    }

    def __init__(self, name, signature=None, altname=None, flags=0, force=False):
        self.name = name
        self.signature = signature
        self.altname = altname
        self.flags = flags
        self.force = force

    @property
    def static(self):
        return self.flags & self.FLAG_STATIC

    @property
    def optional(self):
        return self.flags & self.FLAG_OPTIONAL

    @property
    def restype(self):
        return self.get_types()[0]

    @property
    def argtypes(self):
        return self.get_types()[1]

    @property
    def modifier(self):
        return self.type_modifiers.get(self.restype, "Object")

    @property
    def cname(self):
        name = self.altname or self.name
        return name.replace('$', '_')

    def validate(self):
        """Must have a non-null name and signature"""
        if not self.name:
            raise ValueError("Must have a name")
        if not self.signature:
            raise ValueError("Must have a signature")

    def __eq__(self, other):
        if other.name != self.name:
            return False
        if other.signature and self.signature:
            if other.signature != self.signature:
                return False
        return True

    def _get_type(self, typesig):
        """Parse single type signature to jni type"""
        dimensions = typesig.count('[')
        if dimensions > 1:
            return 'jarray'
        typesig = typesig.lstrip('[')
        typebyte, classname = typesig[0], typesig[1:]
        jnitype = self.types[typebyte]
        if jnitype == 'jobject' and not dimensions:
            if classname == 'java/lang/Class;':
                jnitype = 'jclass'
            elif classname == 'java/lang/String;':
                jnitype = 'jstring'
            elif classname == 'java/lang/Throwable;':
                jnitype = 'jthrowable'
        if jnitype != 'void' and dimensions:
            jnitype += 'Array'
        return jnitype
    
    def _get_types(self):
        """Parse complete signature to jni type(s)"""
        match = self.re_sig.match(self.signature)
        if not match:
            raise ValueError('Bad signature')
        _argtypes, _restype = match.groups()
        argtypes = []
        while _argtypes:
            match = self.re_type.match(_argtypes)
            if not match:
                raise ValueError('Bad signature')
            argtypes.append(self._get_type(match.group(0)))
            _argtypes = _argtypes[match.end():]
        match = self.re_type.match(_restype)
        if not match:
            raise ValueError('Bad signature')
        restype = self._get_type(match.group(0))
        return restype, argtypes

    def get_types(self):
        if hasattr(self, '_types'):
            return self._types
        if not self.signature:
            return None, None
        self._types = self._get_types()
        return self._types

    @classmethod
    def reduce(cls, javap, nji):
        return cls(
            javap.name,
            javap.signature or nji.signature, 
            altname=nji.altname,
            flags=javap.flags | nji.flags,
            force=nji.force
        )


class JniField(JniMember):

    def __init__(self, *args, **kwargs):
        super(JniField, self).__init__(*args, **kwargs)
        if self.argtypes:
            raise ValueError('Field cannot have arguments')

    def __str__(self):
        return '%s%s %s' % (
            'static ' if self.static else '',
            self.name,
            self.signature
        ) 

    def add_to(self, klass):
        klass.add_field(self)

    @property
    def type(self):
        return self.restype


class JniConstructor(JniMember):

    def __str__(self):
        return '%s<init>() %s' % (
            'static ' if self.static else '',
            self.signature
        ) 

    def add_to(self, klass):
        klass.add_constructor(self)


class JniMethod(JniMember):

    def __str__(self):
        return '%s%s() %s' % (
            'static ' if self.static else '',
            self.name,
            self.signature
        ) 

    def add_to(self, klass):
        klass.add_method(self)



import os
import glob
from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name='nji',
    version='0.1.4',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'nji': ['templates/*']},
    scripts=glob.glob('scripts/*'),
    install_requires=['jinja2'],
    include_package_data=True,
    license='BSD License',
    description='Native Java Interface code generator.',
    long_description=README,
    author='greenbender',
    author_email='byron.platt@gmail.com',
    url='https://github.com/greenbender/libnji',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
)



#!/usr/bin/python

from .jni import *
from .javap import *
import json
import os


__all__ = [
    'template_h',
    'template_c',
    'parse',
]


# default locations for templates
_root = os.path.dirname(__file__)
_templates = os.path.join(_root, 'templates')
template_h = os.path.join(_templates, 'default.h')
template_c = os.path.join(_templates, 'default.c')


def _nji_to_class(data):
    """Parse an NJI file and return a JniClass"""

    name = data.pop('class')
    fields = data.pop('fields', [])
    constructors = data.pop('constructors', [])
    methods = data.pop('methods', [])

    clazz = JniClass(name, **data)

    # fields
    if isinstance(fields, list):
        for field in fields:
            name = field.pop('name')
            field.setdefault('flags', 0)
            if field.pop('static', False):
                field['flags'] |= JniField.FLAG_STATIC
            if field.pop('optional', False):
                field['flags'] |= JniField.FLAG_OPTIONAL
            JniField(name, **field).add_to(clazz)
    elif fields == '*':
        clazz.fields = JniAll

    # constructors
    if isinstance(constructors, list):
        for constructor in constructors:
            constructor.setdefault('flags', 0)
            if constructor.pop('optional', False):
                constructor['flags'] |= JniConstructor.FLAG_OPTIONAL
            JniConstructor(clazz.shortname, **constructor).add_to(clazz)
    elif constructors == '*':
        clazz.constructors = JniAll

    # methods
    if isinstance(methods, list):
        for method in methods:
            name = method.pop('name')
            method.setdefault('flags', 0)
            if method.pop('static', False):
                method['flags'] |= JniMethod.FLAG_STATIC
            if method.pop('optional', False):
                method['flags'] |= JniMethod.FLAG_OPTIONAL
            JniMethod(name, **method).add_to(clazz)
    elif methods == '*':
        clazz.methods = JniAll

    return clazz


def _javap_to_class(data, classpath=None):
    """Parse the output of javap for a java class and return a JniClass"""

    classname = data['class']

    # build classpath
    classpath = classpath or []
    extra = data.get('classpath')
    if extra:
        if isinstance(extra, list):
            classpath.extend(extra)
        else:
            classpath.append(extra)

    # parse the complete class
    clazz_full = JavaP.parse_class(classname, classpath)

    # parse the partial class
    clazz_partial = _nji_to_class(data)

    return JniClass.reduce(clazz_full, clazz_partial)


def parse(fd, classpath=None):
    """Parse an NJI file and generate a JniClass based on the NJI file itself
    and any additional information that can be retrieved from javap"""
    
    data = json.load(fd)
    force = data.get('force', False)

    # create class directly
    if force:
        clazz = _nji_to_class(data)

    # create class using javap
    else:
        clazz = _javap_to_class(data, classpath=classpath)
        
    if clazz:
        clazz.validate()
        clazz.uniqify()

    return clazz
